; ==============================================================
;  _ _         __           _   _                  ___   __
; (_) | ___   / / __   __ _| |_(_)_   _____ ___  _( _ ) / /_
; | | |/ _ \ / / '_ \ / _` | __| \ \ / / _ (_) \/ / _ \| '_ \
; | | | (_) / /| | | | (_| | |_| |\ V /  __/_ >  < (_) | (_) |
; |_|_|\___/_/ |_| |_|\__,_|\__|_| \_/ \___(_)_/\_\___/ \___/
; a tiny virtual computer : 64kw RAM, 32-bit, Dual Stack MISC
;                   (c) charles childers
; ==============================================================

MODULEALIGN equ 1<<0
MEMINFO equ 1<<1
FLAGS equ MODULEALIGN | MEMINFO
MAGIC equ 0x1BADB002
CHECKSUM equ -(MAGIC + FLAGS)

section .multiboot
align 4
MultiBootHeader:
  dd MAGIC
  dd FLAGS
  dd CHECKSUM

; ==============================================================

section .text
extern main
extern isr_handler
global _start
global getchar
global idt_flush

STACKSIZE equ 0x4000

align 4
_start:
  lgdt [gdt]
  mov  esp, stack+STACKSIZE
  push dword 0
  push dword 0
  jmp  main
  jmp  $

; ==============================================================

idt_flush:
   mov eax, [esp+4]  ; get the pointer to the IDT, passed as
                     ; a parameter.
   lidt [eax]        ; load the IDT
   ret

isr_common_stub:
   pusha          ; push edi, esi, ebp, esp, ebx, edx, ecx, eax
   mov ax, ds     ; lower 16-bits of eax = ds.
   push eax       ; save the data segment descriptor

   mov ax, 0x10   ; load the kernel data segment descriptor
   mov ds, ax
   mov es, ax
   mov fs, ax
   mov gs, ax

   call isr_handler

   pop eax        ; reload the original data segment descriptor
   mov ds, ax
   mov es, ax
   mov fs, ax
   mov gs, ax

   popa           ; pop edi, esi, ebp...
   add esp, 8     ; clean up the pushed error code and the
                  ; pushed ISR number
   sti            ; enable interrupts again
   iret           ; pop 5 things at once: CS, EIP, EFLAGS,
                  ; SS, and ESP

%macro ISR_GENERIC 1
  [GLOBAL isr%1]
  isr%1:
    cli
    push byte 0
    push byte %1
    jmp isr_common_stub
%endmacro

ISR_GENERIC 0
ISR_GENERIC 1
ISR_GENERIC 2
ISR_GENERIC 3
ISR_GENERIC 4
ISR_GENERIC 5
ISR_GENERIC 6
ISR_GENERIC 7
ISR_GENERIC 8
ISR_GENERIC 9
ISR_GENERIC 10
ISR_GENERIC 11
ISR_GENERIC 12
ISR_GENERIC 13
ISR_GENERIC 14
ISR_GENERIC 15
ISR_GENERIC 16
ISR_GENERIC 17
ISR_GENERIC 18
ISR_GENERIC 19
ISR_GENERIC 20
ISR_GENERIC 21
ISR_GENERIC 22
ISR_GENERIC 23
ISR_GENERIC 24
ISR_GENERIC 25
ISR_GENERIC 26
ISR_GENERIC 27
ISR_GENERIC 28
ISR_GENERIC 29
ISR_GENERIC 30
ISR_GENERIC 31

; ==============================================================


; ==============================================================

align 4
getchar:
    xor     eax, eax               ; Clear eax
.a: in      al, 64h                ; Is any data waiting?
    test    al, 1                  ; Is character = ASCII 0?
    jz      .a                     ; Yes? Try again
    in      al, 60h                ; Otherwise, read scancode
    xor     edx, edx               ; edx: 0=make, 1=break
    test    al, 80h                ; Is character = HEX 80h?
    jz      .b                     ; Skip the next line
    inc     edx                    ; Update edx
.b: and     al, 7Fh                ; Filters to handle ignored keys
    cmp     al, 39h                ; Check if the key is ignored
    ja      .check_fn_keys         ; Check for function keys
    mov     ecx, [board]           ; Load the keymap
    mov     al, [ecx+eax]          ; Get the key ASCII char
    or      al, al                 ; Is it = 0?
    js      .shift                 ; No, use CAPITALS
    jz      .a                     ; Ignore 0's
    or      dl, dl                 ; Filter for break code
    jnz     .a                     ; Ignore break code
    jmp     .done                  ; Skip to end
.shift:
    mov     ecx, [edx*4 + .shifts] ; Load the CAPITAL keymap
    mov     [board], ecx           ; Store into BOARD pointer
    jmp     .a                     ; Try again
.check_fn_keys:
    or      dl, dl                 ; Filter for break code
    jnz     .a                     ; Ignore break code
    cmp     al, 3Bh                ; F1 scancode
    jl      .a                     ; If less than F1 scancode, skip function key handling
    cmp     al, 58h                ; F12 scancode
    jg      .a                     ; If greater than F12 scancode, skip function key handling
    sub     al, 3Bh                ; Convert scancode to index (F1 = 0, F2 = 1, ...)
    inc     al                     ; Increment by 1 to adjust (-1 for F1, -2 for F2, ...)
    neg     eax                    ; Invert sign
.done:
    ret                            ; Return

section .data

.shifts dd shift, alpha

board dd alpha

alpha:
    db 0,27,"1234567890-=",8        ; ASCII characters for keys
    db 9,"qwertyuiop[]",10          ; ASCII characters for keys
    db 0,"asdfghjkl;'`"             ; ASCII characters for keys
    db -1,"\zxcvbnm,./",-1,"*",0,32,-2 ; ASCII characters for keys

shift:
    db 0,27,"!@#$%^&*()_+",8        ; ASCII characters for shifted keys
    db 9,"QWERTYUIOP{}",10          ; ASCII characters for shifted keys
    db 0,'ASDFGHJKL:"~'             ; ASCII characters for shifted keys
    db -1,"|ZXCVBNM<>?",-1,"*",0,32,-2 ; ASCII characters for shifted keys

; ==============================================================

gdt:
        dw      (.end-.dummy)
        dd      .dummy
        dw      0
.dummy: dw      0,0,0,0
        dw      0xffff,0x0000
        db      0x00,0x9a,0xcf,0x00
        dw      0xffff,0x0000
        db      0x00,0x92,0xcf,0x00
.end:

; ==============================================================

section .bss
align 4
stack:
  resb STACKSIZE
stack_ptr:
